from typing import Tuple
from collections import OrderedDict
from bson import Int64


def validate_recursive(obj, val_fn, depth=0):
    """runs the validation function val_fn(key, val, depth) recursively
    on keys and values, throws if the query is not supported."""

    if isinstance(obj, dict):
        for k, v in obj.items():
            val_fn(k, None, depth)
            validate_recursive(v, val_fn, depth + 1)
    elif isinstance(obj, list):
        all(validate_recursive(el, val_fn, depth) for el in obj)
    else:
        val_fn(None, obj, depth)
    return True


class Query(object):
    """The Query class represents a query object, consisting of
    a number of Predicates. Predicates are named tuples of
    the form (column, op, values).
    This class also contains some utility methods, for example
    exporting a query to MQL syntax, intersecting a query
    with an index (for Index Selection), etc.
    """

    def __init__(self):
        self._filter = OrderedDict()
        self._limit = None
        self._projection = None
        self._sort = None

    def _is_filter_supported(self, f):
        """returns whether the filter is supported in this version of mindexer."""

    @property
    def filter(self):
        """returns the filter for this query."""
        return self._filter

    @filter.setter
    def filter(self, f: int):
        """set a filter for this query. None means no filter = {}."""
        assert type(f) == dict, "limit must be a dict"

        # validation for filter object
        def val_fn(k, v, d):
            if k == "$and" and d > 0:
                raise NotImplementedError("$and is only supported at the top level")
            if k in ["$nor", "$or", "$text"]:
                raise NotImplementedError(f"queries with {k} not supported.")

        # this throws if validation fails
        validate_recursive(f, val_fn)

        # filter out $comment fields
        f = {k: v for k, v in f.items() if k != "$comment"}

        # flatten top-level $and to implicit syntax
        if "$and" in f:
            temp_q = Query()
            for el in f["$and"]:
                temp_q.add_predicate(el)
                self._filter = temp_q.filter
            return

        self._filter = f

    @property
    def limit(self):
        """returns the limit for this query."""
        return self._limit

    @limit.setter
    def limit(self, n: int):
        """set a limit for this query. None means no limit."""
        assert isinstance(n, (int, Int64)), f"limit must be an integer, is {type(n)}"
        self._limit = n

    @property
    def projection(self):
        """returns the projection tuple of this query."""
        return self._projection

    @projection.setter
    def projection(self, p: Tuple[str, ...]):
        """set the projection tuple for this query."""
        self._projection = p

    @property
    def sort(self):
        """returns the sort tuple of this query."""
        return self._sort

    @sort.setter
    def sort(self, s: Tuple[str, ...]):
        """set the sort tuple for this query."""
        self._sort = s

    @property
    def fields(self):
        """return all fields of the query, whether they are part of the predicates,
        projection or sort.
        """
        fields = list(self.filter.keys())
        if self.sort:
            fields += list(self.sort)
        if self.projection:
            fields += list(self.projection)
        return sorted(set(fields))

    def add_predicate(self, predicate):
        """adds a predicate to the query."""

        assert isinstance(predicate, dict)

        # get first (and only) key
        key = next(iter(predicate))
        if not key in self.filter:
            self._filter.update(predicate)
        else:
            # get filter and predicate values
            fval = self._filter[key]
            pval = predicate[key]
            if type(fval) == dict and type(pval) == dict:
                fval.update(pval)
            else:
                raise Exception(
                    f"Error: can't update {key} field with predicate {predicate}."
                )

    def add_predicates(self, preds):
        """add multiple predicates to a query at once."""

        assert isinstance(preds, dict)

        for k, v in preds.items():
            self.add_predicate({k: v})

    def index_intersect(self, index):
        """returns a copy of this query that only contains predicates on the
        provided index fields, left to right, up until an index field was
        not included in the query.

        Example: query is {a: true, b: {$lt: 20}, c: 5}
                 query.index_intersect(("b", "d", "c")) return the query
                 equivalent to {b: {$lt: 20}}, because d is not present
                 in the query and aborts the algorithm.

        """
        query = Query()

        for field in index:
            if field in self.filter:
                query.add_predicate({field: self.filter[field]})
            else:
                break

        return query

    def is_subset(self, index):
        """returns true if all predicate fields are included in the index.
        This is necessary, but not sufficient to be a covered by the index.
        It is used to determine if a limit caps the cost of the query or not.
        """
        predicate_fields = set(self.filter.keys())
        return predicate_fields.issubset(set(index))

    def is_covered(self, index):
        """returns true if this query is covered by the index, false otherwise.
        A query without a projection is never covered. A query with projection
        is covered if the union of all predicate fields and projected fields
        appear in the index.
        """
        if self.projection is None:
            return False

        predicate_fields = tuple(self.filter.keys())
        fields_to_cover = set(predicate_fields + self.projection)

        return fields_to_cover.issubset(set(index))

    def can_use_sort(self, index):
        """returns true if this query has a sort and the index can be used to sort,
        false otherwise.
        A query can be sorted by an index if any of the following is true:
            a) the sort fields are the same as the index fields (incl. order)
            b) the sort fields are a prefix sequence of the index fields
            c) the sort fields are a sub-sequence of the index fields and
               the query only uses equality predicates on all fields preceeding
               the sort fields.

               Example: index on ('a', 'b', 'c', 'd')
                        query {a: 5, b: {$gt: 6}} with sort ('b', 'c')
                        The predicate preceeding the sort fields is 'a' and is
                        an equality predicate. The index can be used to sort.
        """
        if self.sort is None:
            # TODO check this: if the query doesn't need a sort, the sort fields sequence
            # is and empty list, and theoretically a prefix of any index.
            return False

        # cover case a) and b)
        if self.sort == index[: len(self.sort)]:
            return True

        # cover case c)
        sub_idx = -1
        for i in range(len(index) - len(self.sort)):
            if self.sort == index[i : i + len(self.sort)]:
                # sort fields are a sub-sequence of index fields
                sub_idx = i

        def is_equality_cmp(field):
            if field not in self.filter:
                # it's not an equality comparison if the field isn't in the query
                return False
            if isinstance(self.filter[field], dict):
                # if none (= not any) of the keys start with $, then it's an equality comparison
                return not any(key.startswith("$") for key in self.filter[field].keys())
            else:
                # if value is not a dictionary, it's always an equality comp.
                return True

        if sub_idx != -1:
            # check for preceeding equality predicates
            if all(is_equality_cmp(key) for key in index[:sub_idx]):
                return True

        return False

    def __repr__(self):
        s = []
        s.append(f"filter={self.filter}")
        if self.sort:
            s.append(f"sort={self.sort}")
        if self.limit:
            s.append(f"limit={self.limit}")
        if self.projection:
            s.append(f"projection={self.projection}")
        return f"Query({', '.join(s)})"

    def __len__(self):
        """The length of a query, using 'len(query)', is the number of
        predicates it contains."""
        return len(self.filter.keys())

    def to_mql(self):
        """converts the query to MQL syntax."""
        return self.filter

    @classmethod
    def from_mql(cls, query):
        """converts MQL syntax to a Query object."""
        q = cls()
        if query:
            q.filter = query

        return q



import unittest
import pandas as pd
import numpy as np
from mindexer.utils.query import Query, validate_recursive


class TestQuery(unittest.TestCase):
    def test_init_empty(self):
        query = Query()
        self.assertEqual(query.filter, {})

    def test_init_from_mql(self):
        query = Query.from_mql({"foo": {"$gt": 16}})
        self.assertEqual(query.filter, {"foo": {"$gt": 16}})
        self.assertIsNone(query.limit)
        self.assertIsNone(query.projection)
        self.assertIsNone(query.sort)

    def test_limit(self):
        query = Query()
        query.limit = 12
        self.assertEqual(query.limit, 12)

        def assign_float():
            query.limit = 12.5

        self.assertRaises(AssertionError, assign_float)

    def test_projection(self):
        query = Query()
        query.projection = ("foo", "bar")
        self.assertEqual(query.projection, ("foo", "bar"))

    def test_fields(self):
        query = Query.from_mql({"foo": {"$gt": 16}, "bar": {"$in": [1, 2, 3]}})
        query.sort = ("a", "foo")
        query.projection = ("foo", "new")

        self.assertEqual(query.fields, ["a", "bar", "foo", "new"])

    def test_add_predicate_new(self):
        query = Query.from_mql({"foo": {"$gt": 16}, "bar": {"$in": [1, 2, 3]}})
        query.add_predicate({"woo": "hoo"})

        self.assertTrue("woo" in query.fields)
        self.assertEqual(query.filter["woo"], "hoo")

    def test_add_predicate_existing(self):
        query = Query.from_mql({"foo": {"$gt": 16}, "bar": {"$in": [1, 2, 3]}})
        query.add_predicate({"foo": {"$lt": 20}})

        self.assertEqual(query.filter["foo"], {"$gt": 16, "$lt": 20})

    def test_add_predicate_conflict(self):
        query = Query.from_mql({"foo": 16, "bar": {"$in": [1, 2, 3]}})

        def fn():
            query.add_predicate({"foo": {"$lt": 20}})

        self.assertRaisesRegex(Exception, "can't update foo", fn)

    def test_add_predicates(self):
        query = Query.from_mql({"foo": {"$gt": 5}})
        query.add_predicates({"foo": {"$lt": 10}, "bar": True})

        self.assertEqual(query.filter, {"foo": {"$gt": 5, "$lt": 10}, "bar": True})

    def test_index_intersect(self):
        pass

    def test_covered_no_projection(self):
        query = Query()
        self.assertFalse(query.is_covered(("Make", "Unladen Weight")))

    def test_covered_projection_no_predicates(self):
        query = Query()
        query.projection = ("Make",)
        self.assertFalse(query.is_covered(("City", "State")))
        self.assertTrue(query.is_covered(("City", "Make")))
        self.assertTrue(query.is_covered(("Make",)))

    def test_covered_projection_predicates(self):
        query = Query.from_mql(
            {"Unladen Weight": 2000, "Make": {"$in": ["INFIN", "HYUND"]}}
        )

        query.projection = ("City",)
        self.assertFalse(query.is_covered(("City", "State", "Make")))
        self.assertTrue(query.is_covered(("City", "Make", "Unladen Weight")))
        self.assertTrue(query.is_covered(("Make", "State", "City", "Unladen Weight")))

    def test_subset(self):
        query = Query.from_mql(
            {"Unladen Weight": 2000, "Make": {"$in": ["INFIN", "HYUND"]}}
        )

        self.assertFalse(query.is_subset(("City", "State", "Make")))
        self.assertTrue(query.is_subset(("Make", "Unladen Weight")))
        self.assertTrue(query.is_subset(("Make", "State", "City", "Unladen Weight")))

    def test_sort(self):
        query = Query()
        query.sort = ("foo", "bar")
        self.assertEqual(query.sort, ("foo", "bar"))

    def test_can_use_sort_equal(self):
        query = Query()
        query.sort = ("foo", "bar")
        self.assertTrue(query.can_use_sort(("foo", "bar")))

    def test_can_use_sort_prefix(self):
        query = Query()
        query.sort = ("foo", "bar")
        self.assertTrue(query.can_use_sort(("foo", "bar", "baz")))

    def test_can_use_sort_no_prefix(self):
        query = Query()
        query.sort = ("foo", "bar")
        self.assertFalse(query.can_use_sort(("foo", "other", "blah")))

    def test_can_use_sort_sub_seq_preceeding_eq(self):
        query = Query.from_mql(
            {"Unladen Weight": 2000, "Make": {"$in": ["INFIN", "HYUND"]}}
        )
        query.sort = ("Make", "City")
        self.assertTrue(query.can_use_sort(("Unladen Weight", "Make", "City", "State")))

    def test_can_use_sort_linkbench_1(self):
        query = Query.from_mql(
            {'id1': 38020, 'link_type': 123456790, 'time': {'$gte': 0, '$lte': 9223372036854775807}, 'visibility': 1}
        )
        query.sort = ("time",)

        self.assertTrue(
            query.can_use_sort(('id1', 'link_type', 'visibility', 'time', 'id2', 'version', 'data'))
        )


    def test_can_use_sort_sub_seq_preceeding_not_eq(self):
        query = Query.from_mql(
            {"Unladen Weight": {"$gt": 2000}, "Make": {"$in": ["INFIN", "HYUND"]}}
        )
        query.sort = ("Make", "City")
        self.assertFalse(
            query.can_use_sort(("Unladen Weight", "Make", "City", "State"))
        )

    def test_fields(self):
        query = Query.from_mql(
            {"Unladen Weight": 2000, "Make": {"$in": ["INFIN", "HYUND"]}}
        )

        query.projection = ("City",)
        query.sort = ("State", "County")

        self.assertEqual(
            query.fields, ["City", "County", "Make", "State", "Unladen Weight"]
        )

    def test_to_mql_dup_keys(self):
        query = Query.from_mql(
            {
                "Unladen Weight": {"$gte": 1300, "$lt": 2000},
                "Make": {"$in": ["INFIN", "HYUND"]},
            }
        )

        mql = query.to_mql()
        self.assertEqual(
            mql,
            {
                "Unladen Weight": {"$gte": 1300, "$lt": 2000},
                "Make": {"$in": ["INFIN", "HYUND"]},
            },
        )

    def test_query_index_intersect_prefix(self):
        query = Query.from_mql({"Unladen Weight": {"$lte": 3700}, "Make": "HYUND"})

        intersected = query.index_intersect(("Unladen Weight", "Foo Field"))
        self.assertEqual(intersected.fields, ["Unladen Weight"])

    def test_query_index_intersect_empty(self):
        query = Query.from_mql({"Unladen Weight": {"$lte": 3700}, "Make": "HYUND"})

        intersected = query.index_intersect(("Foo Field", "Unladen Weight"))
        self.assertEqual(intersected.fields, [])

    def test_query_index_intersect_multi(self):
        query = Query.from_mql(
            {"Unladen Weight": {"$lte": 3700}, "Make": "HYUND", "State": "NY"}
        )

        intersected = query.index_intersect(
            ("State", "Unladen Weight", "Foo Field", "Make")
        )
        self.assertEqual(intersected.fields, ["State", "Unladen Weight"])

    def test_query_with_id(self):
        query = Query.from_mql(
            {
                "_id": "62b50910",
                "tenantId": "4be48469",
                "$comment": "62b55590c",
            }
        )
        self.assertEqual(query.fields, ["_id", "tenantId"])

    def test_validation_pass(self):
        Query.from_mql({"_id": 1, "foo": {"$gt": 20, "$lte": 30}, "bar": True})

    def test_validation_or_0(self):
        def fn():
            Query.from_mql({"$or": [{"_id": 1}, {"foo": {"$gt": 20, "$lte": 30}}]})

        self.assertRaisesRegex(
            NotImplementedError, "queries with \$or not supported", fn
        )

    def test_validation_and_0(self):
        query = Query.from_mql({"$and": [{"_id": 1}, {"foo": {"$gt": 20, "$lte": 30}}]})
        # flatten into implicit $and query
        self.assertEqual(query.filter, {"_id": 1, "foo": {"$gt": 20, "$lte": 30}})

    def test_validation_and_explicit_to_implicit(self):
        query = Query.from_mql({"$and": [{"foo": {"$gt": 20}}, {"foo": {"$lte": 30}}]})
        # flatten into implicit $and query
        self.assertEqual(query.filter, {"foo": {"$gt": 20, "$lte": 30}})

    def test_validation_and_1(self):
        def fn():
            Query.from_mql(
                {
                    "$and": [
                        {"$and": [{"foo": 1}, {"bar": 1}]},
                        {"foo": {"$gt": 20, "$lte": 30}},
                    ]
                }
            )

        self.assertRaisesRegex(
            NotImplementedError, "\$and is only supported at the top level", fn
        )

    def test_validation_or_1(self):
        def fn():
            Query.from_mql(
                {
                    "$and": [
                        {"$or": [{"foo": 1}, {"bar": 1}]},
                        {"foo": {"$gt": 20, "$lte": 30}},
                    ]
                }
            )

        self.assertRaisesRegex(
            NotImplementedError, "queries with \$or not supported", fn
        )


if __name__ == "__main__":
    unittest.main()



import unittest
from unittest.mock import Mock
from mindexer.utils.query import Query
from mindexer.utils.sampling import SampleEstimator


class TestSampleEstimator(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.mongoCollection = Mock()

    def test_pipeline_defaults(self):
        # test a pipeline built with default parameters
        TestSampleEstimator.mongoCollection.count = 100

        query = Query.from_mql({"County": "KINGS", "Model Year": {"$gt": 2010}})

        sample_est = SampleEstimator(TestSampleEstimator.mongoCollection)

        pipeline = sample_est.make_pipeline(query)

        self.assertEqual(
            pipeline,
            [
                {"$match": {"County": "KINGS", "Model Year": {"$gt": 2010}}},
                {"$count": "total"},
            ],
        )

    def test_pipeline_sample_size(self):
        # test building a pipeline when given a sample size
        TestSampleEstimator.mongoCollection.count = 10

        query = Query.from_mql({"County": "KINGS", "Model Year": {"$gt": 2010}})
        sample_est = SampleEstimator(TestSampleEstimator.mongoCollection, sample_size=3)

        pipeline = sample_est.make_pipeline(query)

        self.assertEqual(
            pipeline,
            [
                {"$sample": {"size": 3}},
                {"$match": {"County": "KINGS", "Model Year": {"$gt": 2010}}},
                {"$count": "total"},
            ],
        )

    def test_pipeline_sample_ratio(self):
        # test building a pipelind when given a sample ratio
        TestSampleEstimator.mongoCollection.count = 100

        query = Query.from_mql({"County": "KINGS", "Model Year": {"$gt": 2010}})

        sample_est = SampleEstimator(
            TestSampleEstimator.mongoCollection, sample_ratio=0.5
        )

        pipeline = sample_est.make_pipeline(query)

        self.assertEqual(
            pipeline,
            [
                {"$sample": {"size": 50}},
                {"$match": {"County": "KINGS", "Model Year": {"$gt": 2010}}},
                {"$count": "total"},
            ],
        )

    def test_pipeline_zero(self):
        # test that giving a size of 0 returns error
        TestSampleEstimator.mongoCollection.count = 100

        with self.assertRaises(AssertionError) as err:
            SampleEstimator(TestSampleEstimator.mongoCollection, sample_size=0)
            print(err)

    def test_pipeline_large(self):
        # test that there is an error if given a sample size larger than collection size
        TestSampleEstimator.mongoCollection.count = 100

        with self.assertRaises(AssertionError) as err:
            SampleEstimator(TestSampleEstimator.mongoCollection, sample_size=1000)
            print(err)

    def test_pipeline_round_sample_ratio(self):
        # test that giving a very small sample ratio will throw an error
        TestSampleEstimator.mongoCollection.count = 100

        with self.assertRaises(AssertionError) as err:
            SampleEstimator(TestSampleEstimator.mongoCollection, sample_ratio=0.0001)
            print(err)

    def test_pipeline_with_limit(self):
        # test building a pipeline with a limit on number of documents
        TestSampleEstimator.mongoCollection.count = 100

        query = Query.from_mql({"County": "KINGS", "Model Year": {"$gt": 2010}})

        sample_est = SampleEstimator(
            TestSampleEstimator.mongoCollection, numrows=10, sample_size=3
        )

        pipeline = sample_est.make_pipeline(query)

        self.assertEqual(
            pipeline,
            [
                {"$limit": 10},
                {"$sample": {"size": 3}},
                {"$match": {"County": "KINGS", "Model Year": {"$gt": 2010}}},
                {"$count": "total"},
            ],
        )


if __name__ == "__main__":
    unittest.main()



from setuptools import setup, find_packages


def readme():
    with open("README.md") as f:
        return f.read()


setup(
    name="mindexer",
    version="0.3.1",
    description="Experimental Index Recommendation Tool for MongoDB",
    long_description=readme(),
    long_description_content_type="text/markdown",
    keywords="mongodb indexes recommendations",
    url="http://github.com/mongodb-labs/mindexer",
    author="Thomas Rueckstiess",
    author_email="thomas.rueckstiess@mongodb.com",
    license="MIT",
    packages=find_packages(),
    scripts=["bin/mindexer"],
    install_requires=["numpy", "pandas", "pymongo"],
    zip_safe=False,
)
